@@ -115,6 +115,13 @@ struct perf__doh_socket {
     bool            http2_is_dns;
 
     struct _doh_stats stats;
+
+    char   recvbuf[TCP_RECV_BUF_SIZE];
+    size_t recv_at;
+
+    size_t       num_queries_per_conn, nqpc_timeout;
+    unsigned int nqpc_sent, nqpc_recv;
+    uint64_t     nqpc_ts;
 };
 
 static pthread_mutex_t            _nghttp2_lock      = PTHREAD_MUTEX_INITIALIZER;
@@ -154,6 +161,10 @@ static void perf__doh_connect(struct perf_net_socket* sock)
 {
     int ret;
 
+    self->nqpc_sent = 0;
+    ck_pr_store_uint(&self->nqpc_recv, 0);
+    self->nqpc_ts = 0;
+
     nghttp2_session_del(self->http2.session);
     ret = nghttp2_session_client_new2(&self->http2.session, _nghttp2_callbacks, sock, _nghttp2_option);
     if (ret < 0) {
@@ -291,50 +302,62 @@ static ssize_t perf__doh_recv(struct perf_net_socket* sock, void* buf, size_t le
         return -1;
     }
 
-    uint8_t recvbuf[TCP_RECV_BUF_SIZE];
-    ssize_t n = SSL_read(self->ssl, recvbuf, TCP_RECV_BUF_SIZE);
-    if (!n) {
-        perf__doh_reconnect(sock);
-        PERF_UNLOCK(&self->lock);
-        errno = EAGAIN;
-        return -1;
-    }
-    if (n < 0) {
-        int err = SSL_get_error(self->ssl, n);
-        switch (err) {
-        case SSL_ERROR_WANT_READ:
+    if (self->recv_at < sizeof(self->recvbuf)) {
+        // try to recv more if we can
+        ssize_t n = SSL_read(self->ssl, self->recvbuf + self->recv_at, sizeof(self->recvbuf) - self->recv_at);
+        if (!n) {
+            perf__doh_reconnect(sock);
+            PERF_UNLOCK(&self->lock);
             errno = EAGAIN;
-            break;
-        case SSL_ERROR_SYSCALL:
-            switch (errno) {
-            case ECONNREFUSED:
-            case ECONNRESET:
-            case ENOTCONN:
-                perf__doh_reconnect(sock);
+            return -1;
+        }
+        if (n < 0) {
+            int err = SSL_get_error(self->ssl, n);
+            switch (err) {
+            case SSL_ERROR_WANT_READ:
+                if (self->recv_at) {
+                    // did not recv but we got something in buffer
+                    break;
+                }
                 errno = EAGAIN;
-                break;
+                PERF_UNLOCK(&self->lock);
+                return -1;
+            case SSL_ERROR_SYSCALL:
+                switch (errno) {
+                case ECONNREFUSED:
+                case ECONNRESET:
+                case ENOTCONN:
+                    perf__doh_reconnect(sock);
+                    errno = EAGAIN;
+                    break;
+                default:
+                    break;
+                }
+                PERF_UNLOCK(&self->lock);
+                return -1;
             default:
-                break;
+                errno = EBADF;
+                PERF_UNLOCK(&self->lock);
+                return -1;
             }
-            break;
-        default:
-            errno = EBADF;
-            break;
+        } else {
+            self->recv_at += n;
         }
-        PERF_UNLOCK(&self->lock);
-        return -1;
     }
 
     // this will be processed by nghttp2 callbacks
-    int ret = nghttp2_session_mem_recv(self->http2.session, recvbuf, n);
+    ssize_t ret = nghttp2_session_mem_recv(self->http2.session, (uint8_t*)self->recvbuf, self->recv_at);
     if (ret < 0) {
         perf_log_warning("nghttp2_session_mem_recv failed: %s", nghttp2_strerror(ret));
         PERF_UNLOCK(&self->lock);
         return -1;
     }
-    // TODO: handle partial mem_recv
-    if (ret != n) {
-        perf_log_fatal("perf__doh_recv() mem_recv did not take all");
+    if (ret != self->recv_at) {
+        // only process one response per call
+        memmove(self->recvbuf, self->recvbuf + ret, self->recv_at - ret);
+        self->recv_at -= ret;
+    } else {
+        self->recv_at = 0;
     }
 
     // TODO: is this faster then mem_recv?
@@ -379,6 +402,9 @@ static ssize_t perf__doh_recv(struct perf_net_socket* sock, void* buf, size_t le
 
         // self->have_more = false; TODO
         PERF_UNLOCK(&self->lock);
+        if (self->num_queries_per_conn) {
+            ck_pr_inc_uint(&self->nqpc_recv);
+        }
         return len;
     }
 
@@ -571,6 +597,8 @@ static ssize_t perf__doh_sendto(struct perf_net_socket* sock, uint16_t qid, cons
     }
     PERF_UNLOCK(&self->lock);
 
+    self->nqpc_sent++;
+
     return len;
 }
 
@@ -613,6 +641,21 @@ static int perf__doh_sockready(struct perf_net_socket* sock, int pipe_fd, int64_
             }
             self->is_sending = false;
             sent             = true;
+            self->nqpc_sent++;
+        }
+        if (self->num_queries_per_conn && self->nqpc_sent >= self->num_queries_per_conn) {
+            if (!self->nqpc_ts) {
+                self->nqpc_ts = perf_get_time() + self->nqpc_timeout;
+            }
+            unsigned int r = ck_pr_load_uint(&self->nqpc_recv);
+            if (r >= self->nqpc_sent || perf_get_time() > self->nqpc_ts) {
+                self->do_reconnect = true;
+            }
+            PERF_UNLOCK(&self->lock);
+            if (sent && sock->sent) {
+                sock->sent(sock, self->qid);
+            }
+            return 0;
         }
         PERF_UNLOCK(&self->lock);
         if (sent && sock->sent) {
@@ -850,7 +893,9 @@ static int _http2_data_chunk_recv_cb(nghttp2_session* session,
     }
 
     if (self->http2.dnsmsg_completed) {
-        perf_log_fatal("_http2_data_chunk_recv_cb: chunk received when already having a dns msg");
+        // pause for now and don't process this chunk
+        // can only do one response at a time in perf__doh_recv()
+        return NGHTTP2_ERR_PAUSE;
     }
 
     // TODO: point of nghttp2_session_get_stream_user_data() code?
@@ -881,6 +926,12 @@ static int select_next_proto_cb(SSL* ssl, unsigned char** out,
 }
 #endif /* !OPENSSL_NO_NEXTPROTONEG */
 
+static void perf__doh_num_queries_per_conn(struct perf_net_socket* sock, size_t num_queries_per_conn, size_t timeout)
+{
+    self->num_queries_per_conn = num_queries_per_conn;
+    self->nqpc_timeout         = timeout;
+}
+
 struct perf_net_socket* perf_net_doh_opensocket(const perf_sockaddr_t* server, const perf_sockaddr_t* local, size_t bufsize, void* data, perf_net_sent_cb_t sent, perf_net_event_cb_t event)
 {
     struct perf__doh_socket* tmp  = calloc(1, sizeof(struct perf__doh_socket)); // clang scan-build
@@ -898,6 +949,8 @@ struct perf_net_socket* perf_net_doh_opensocket(const perf_sockaddr_t* server, c
     sock->sockready = perf__doh_sockready;
     sock->have_more = perf__doh_have_more;
 
+    sock->num_queries_per_conn = perf__doh_num_queries_per_conn;
+
     sock->data  = data;
     sock->sent  = sent;
     sock->event = event;
